unit BinaryDD;

// nacte popis BDD, konvertuje na soucet soucinu a odstrani redundance
//
// popis BDD je v textovem souboru, !!! nerozlisuji se mala a velka pismena
// konec souboru nebo radek zacinajici slovem ENDBDD ukonci cteni
// * Radek obsahuje jeden uceleny popis rozhodovaciho uzlu BDD.
//
// * Popis uzlu ma formu:
//     jmeno uzlu, jmeno podminky, spoj_po_nesplneni, spoj_po_splneni_podminky
//   kde spoj je ma nekterou z nasledujicich podob:
//     1)    ANO (nebo YES nebo True nebo T)
//     2)    NE  (nebo NO nebo False nebo F)
//     3)    jmeno navazujiciho uzlu
//     4)    NIC  (pro pripad, kdy z uzlu vede jen jediny funkcni spoj)
//
// *  oddelovaci carky zavazne nejsou, staci mezera nebo jiny spec. znak.
// * Jmeno uzlu i jmeno podminky se sklada vyhradne z pismen a cislic !!!
//   Jmeno obsahuje 1 az max. 8 znaku, a je odlisne od typu ANO i NE i NIC.
//   (Delsi jmeno bude signalizovano jako chyba)
// * Jmena podminek se v ruznych uzlech BDD mohou podle potreby opakovat,
//   naproti tomu kazda definice uzlu musi obsahovat jedinecne jmeno.
// * Mezi seznamem jmen uzlu a seznamem jmen funkci obecne neni zadny vztah.
// * Jmena navaznych uzlu se v ramci definice uzlu nesmi o
//   rovnat jmenu uzlu (tzn. uzel nesmi vest sam do sebe).
// * Popis BDD musi obsahovat aspon jeden spoj typu ANO (YES, T nebo TRUE)
// * Prohledavani zacne od uzlu, do nehoz nevede zadny spoj, a hledaji se
//   cesty ke vsem dostupnym koncum typu ANO, pokud na konci popisu neni
//   uveden radek typu:
//         ENDBDD [jmeno_koncoveho_stavu[,jmeno_vychoziho_uzlu]]
//
// Priklad: logicky soucet 2 podminek, A NEBO B, by mohl mit BDD zadan jako:
//                 1 = A : 2, ANO
//	               2 = B : NE, ANO

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, Math;

procedure VystrahaSyntax;
procedure VystrahaPrazdnySoubor;
procedure VystrahaChybnaStruktura;
procedure VystrahaPreplneni;
procedure ZpravaVypocetUkoncen;
procedure ZpracujRadekPopisuBDD;
function kodJmenaUzlu(jmeno: string): integer;
function kodPodminky(jmeno: string): integer;
procedure ProchazeniUzly;
procedure zapsaniTrasy;
procedure OdstraneniRedundantnichCasti;
procedure VypsatVystupy;

const
  maxVystupZnaku = 60; maxSOP = 64000;
  maxJmen = 4010; maxPocetUzlu = 4000; maxZnakuJmena = 7;
  kANO = 1; kNE = 5; kENDBDD = 9; i32 = 32; bdebug = true;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    Button2: TButton;
    Button1: TButton;
    Button5: TButton;
    OpenDialog1: TOpenDialog;
    Button3: TButton;
    Memo1: TMemo;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Button5Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);

  private

      { Private declarations }
  public
      { Public declarations }
       nazevFormulare: string;
    procedure NactiPopisBDD; // BDD = Binary Decision Diagram
  end;
var
  Form1: TForm1;
  BDDFile, BDDMsg, SOPFile: TextFile;
   
  uzly0, podminky0, uzly, podminky: array[0..maxJmen] of string;
  nUzly, nBC: integer; // aktualni obsazeni poli uzly, resp. podminky
  jDef, jIn, jPlay, jStavy: array[0..maxJmen] of integer;

  radekTextu: string;
  jmenoU, jmenoC, jmenoF, jmenoT: string; //ctene polozky zadani uzlu
  jUzel, jBC, jT, jF: array[0..maxPocetUzlu] of integer;
  idef, iBC, iuF, iuT: integer; // integer kod polozek popisu uzlu
  nuc: integer; // aktualni obsazeni poli jUzel, jBC, jT, jF  (=pocet uzlu)
  ist, status, kstrukt, kpreplneni: integer;  // stav diagnostiky
  ktop, kdef, kend: integer; // kod vychoziho uzlu a koncoveho stavu
  jZasobnikTrasy: array[1..maxPocetUzlu] of integer;
  nZasobnik, kStavy: integer; // obsazeni zasobniku trasy
  maxLevel, maxPlay: integer;
  jSOP, jPrac, jVel: array[1..maxSOP] of integer; // ulozeni vsech soucinu
  nSOP, nSOPCelk, nPrac: integer; // aktualni obsazeni poli
  iSOP, iPrac: integer; // obsah zpracovavane polozky
  taskOK, konecBDD, zadaniOdDo: boolean;
  maska, komplementMaska: array[1..i32] of integer; // bitov maska
  iw, iww: integer; //pocet slov na bitove zakodovani 1 soucinu promennych

  vstupniSoubor,vystupniSoubor: string;

implementation

{$R *.dfm}

procedure TForm1.NactiPopisBDD;
var
  buf, BackupName: string;
  i, ii: integer;
begin
   // iniciace rezervovanych jmen
  uzly[0] := 'NIC     '; uzly[1] := 'ANO     '; uzly[2] := 'YES     ';
  uzly[3] := 'TRUE    '; uzly[4] := 'T       '; uzly[5] := 'NE      ';
  uzly[6] := 'NO      '; uzly[7] := 'FALSE   '; uzly[8] := 'F       ';
  uzly[9] := 'ENDBDD  ';
   // iniciace priznaku chyb
  ist := 0; status := 0; kstrukt := 0; kpreplneni := 0;
   //  iniciace promennych ukazujicich obsazeni pracovnich poli
  nUzly := -1; nBC := 0; nuc := -1; nZasobnik := 0;
   // iniciace priznaku konce zadani popisu BDD
  konecBDD := False; zadaniOdDo := False; ktop := 0; kend := 0;
   // iniciace pomocnych poli
  for i := 0 to maxJmen do
  begin
    jDef[i] := -1;
    jIn[i] := 0;
    jStavy[i] := 0;
  end;
  ii := 1;
  for i := 1 to i32 - 1 do
  begin
    maska[i] := ii;
    komplementMaska[i] := not ii;
    ii := ii + ii;
  end;
  maska[i32] := ii;
  komplementMaska[i32] := not ii;
   //  zda byl otevren soubor obsahujc popis uzlu BDD
  if OpenDialog1.Execute then { Display Open dialog box }
  begin
    if not(OpenDialog1.FileName='') then
    begin
     vstupniSoubor:= OpenDialog1.FileName;
     AssignFile(BDDFile, vstupniSoubor); { File selected in dialog }
     Reset(BDDFile);
     form1.Caption:=nazevFormulare+' - '+ vstupniSoubor;

  buf := OpenDialog1.FileName;
   // zalozi vystupni soubor pro kontrolni vypis zadani
  vystupniSoubor :=   buf + '.Out';
   //vystupniSoubor := 'Out.txt';
  if FileExists(vystupniSoubor) then
  begin
    BackupName := ChangeFileExt(BackupName, '.BAK');
  end;
  AssignFile(BDDMsg, vystupniSoubor);
  Rewrite(BDDMsg);
  Writeln(BDDMsg,'Kontrolni vypis zadani BDD s ocislovanymi radky'
  + #13 + #10 + 'vyznam = jmeno uzlu, podminka,' +
  ' vystup_pro_NE, vystup_pro_ANO :');

  while not Eof(BDDFile) do
  begin
    try
      ReadLn(BDDFile, radekTextu);
      if not (radekTextu = '') then
      begin
        ist := ist + 1;
            // vypise cislovany radek vstupu
        buf := ' ' + inttostr(ist) + ': ' + radekTextu;
            //FileWrite(FileHandle, buf, length(buf));
        Writeln(BDDMsg, buf);
        ZpracujRadekPopisuBDD;
      end;
    finally
       // CloseFile(BDDFile); // uzave soubory
       // CloseFile(BDDMsg); // uzave soubory
    end;
  end;
  if (ist = 0) then status := -1;
  taskOK := (status = 0) and (kpreplneni = 0) and (kstrukt = 0);
  if taskOK then
  begin
    ProchazeniUzly;
    if ((kstrukt = 0) and (status = 0)) then
    begin
      if (bdebug) then
      begin
        Writeln(BDDMsg, 'Kontrolni vypis - nalezene souciny:');
        VypsatVystupy;
      end;
      OdstraneniRedundantnichCasti;
      Writeln(BDDMsg,'Vysledek - neredundantni souciny, uvedene vzestupne:');
      VypsatVystupy;
      ZpravaVypocetUkoncen;
      /////////////////
      if not(vstupniSoubor='') then
      begin
       CloseFile(BDDFile); // uzave soubory
       CloseFile(BDDMsg); // uzave soubory
      end;

    end;
  end
  else begin
    if (status = -1) then VystrahaPrazdnySoubor;
    if (status > 0) then VystrahaSyntax;
    if (kpreplneni > 0) then VystrahaPreplneni;
    if (kstrukt > 0) then VystrahaChybnaStruktura;

   end;
  end;

  end;
end;
///*********************************************************************

procedure ZpracujRadekPopisuBDD;
var
  i, n, k, m, delkaRadku, kUzly, kBC: Integer;
  radekUp, slovo : string;
  retez: array[0..3] of string;
  alfanum, oddel, uzelBezChyb, uzZapsany: boolean;
begin
  uzelBezChyb := True;
   // odstrani pocatecni a koncove mezery a prevede na velka pismena
  radekTextu := trim(radekTextu);
  radekUp := AnsiUpperCase(radekTextu);
  delkaRadku := Length(radekUp)+1;
  radekUp := radekUp + ' ';
  for m := 0 to 3 do retez[m] := '        ';
   // priprava nacitani ctyr polozek popisu uzlu v cyklu pro m od 0 do 3
  m := 0; // poradove cislo polozky zleva
  i := 0; // cislo zpracovavaneho znaku, i jde od 1 do delkaRadku
  k := 0; // citac poctu znaku ve slove, max k=8, doplni se mezery zprava
  slovo := ''; // buffer pro pripojovani znaku, nyni prazdny
  oddel := True; // zacina se stavem, kdy se zatim nezapisuje do slova
                   // hlavni cyklus zpracovani zadani uzlu, nacteni 4 slov (alfanumer.jmen)
  while (m < 4) do
  begin
    if (i >= delkaRadku) then break;
    i := i + 1;
    n := Integer(radekUp[i]);
    alfanum := ((n >= 48) and (n <= 57)) or ((n >= 65) and (n <= 90));
    if (alfanum) then
    begin // nalezen dalsi alfanumericky znak
      if (k < 8) then
      begin // pripoji dalsi alfanumericky znak do bufferu;
        oddel := False;
        k := k + 1;
        slovo := slovo + radekTextu[i];
      end
      else begin // jmeno m vice nez 8 znaku - je to chyba
        uzelBezChyb := False;
        status := 108;
        writeln(BDDMsg,'jmeno ma vice nez 8 znaku')
      end;
    end
    else begin // nalezen oddelovaci znak
      if not (oddel) then
      begin // doplneni mezer, zaznam jmena a vyprazdneni bufferu
        while (k < 8) do
        begin
          k := k + 1;
          slovo := slovo + ' ';
        end;
        retez[m] := slovo;
        m := m + 1;
        slovo := '';
        k := 0;
      end;
      oddel := True;
    end;
  end;
   // jmeno noveho uzlu
  kUzly := nUzly;
  kBC := nBC;
  jmenoU := AnsiUpperCase(retez[0]);
  idef := kodJmenaUzlu(jmenoU);
  zadaniOdDo := (idef = kENDBDD);
  if ((m < 3) and not (zadaniOdDo)) then
  begin
    uzelBezChyb := False;
    status := 104;
    writeln(BDDMsg,'popis uzlu ma mene nez 4 jmena')
  end;
  if (uzelBezChyb) then
  begin
    uzZapsany := ((kUzly = nUzly) and (jDef[idef] > 0));
    if (uzZapsany or (idef <= kENDBDD)) then
    begin // nove jmeno nebylo zapsano
      if (uzZapsany) then
      begin
        kstrukt := 2; // takovy uzel uz byl zadan
        writeln(BDDMsg,'uzel tohoto jmena uz byl definovan')
      end;
      if (idef < kENDBDD) then
      begin           // uzel ma preddefinovene jmeno
        kstrukt := 3; // jmenem uzlu je rezervovane jmeno koncoveho stavu
        writeln(BDDMsg,'uzel ma nedovolene jmeno')
      end;
    end
    else begin // uzel tohoto jmena dosud nebyl zadan
      if ((nuc < maxPocetUzlu) and (idef > kENDBDD)) then
      begin // zapis definice uzlu do pracovnich jmen a poli
        nuc := nuc + 1;
        jUzel[nuc] := idef; // kd uzlu (a taky adresa jmena v Uzly)
        jDef[idef] := nuc; // adresa, kde pro kod uzlu lezi jeho definice
        uzly0[idef] := retez[0];
      end
      else begin // prilis mnoho jmen
        kpreplneni := maxPocetUzlu;
        writeln(BDDMsg,'prilis mnoho podminek');
      end;
      // jmeno logicke podminky spojene s novym uzlem
      jmenoC := AnsiUpperCase(retez[1]);
      iBC := kodPodminky(jmenoC);
      jBC[nuc] := iBC;
      if (kBC < nBC) then podminky0[nBC] := retez[1];
      // jmeno navazneho uzlu/koncoveho stavu, neplati-li podminka
      jmenoF := AnsiUpperCase(retez[2]);
      iuF := kodJmenaUzlu(jmenoF);
      jF[nuc] := iuF; // kod nasledujiciho uzlu pri nesplneni podminky
      jIn[iuF] := jIn[iuF] + 1; // scita se kolikrat je vstup do uzlu
      // jmeno navazneho uzlu/koncoveho stavu, je-li podminka splnena
      jmenoT := AnsiUpperCase(retez[3]);
      iuT := kodJmenaUzlu(jmenoT);
      jT[nuc] := iuT; // kod nasledujiciho uzlu pri splneni podminky
      jIn[iuT] := jIn[iuT] + 1; // scita se kolikrat je vstup do uzlu
      if ((iuF = idef) or (iuT = idef)) then
      begin
        kstrukt := 1;
        writeln(BDDMsg,'uzel vede sam do sebe')
      end;
    end; // konec pro: uzel tohoto jmena nebyl dosud zadan
    if (zadaniOdDo) then
    begin // radek neni popisem uzlu ale zadanim, odkud kam hledat
      if (m > 1) then
      begin
        kend := kodJmenaUzlu(AnsiUpperCase(retez[1]));
        if ((kend >= kENDBDD) or (kend < 1)) then
        begin
          status := 105;
          writeln(BDDMsg,'nevhodne jmeno koncoveho stavu')
        end;
      end;
      if (m > 2) then
      begin
        kUzly := nUzly;
        ktop := kodJmenaUzlu(AnsiUpperCase(retez[2]));
        if (ktop <= kENDBDD) then
        begin
          status := 106;
          writeln(BDDMsg,'nevhodne jmeno vychoziho uzlu')
        end;
        if (kUzly < nUzly) then
        begin
          status := 107;
          writeln(BDDMsg,'nezname jmeno vychoziho uzlu')
        end;
      end;
    end;
  end;
end;
///**************************************************************

function kodJmenaUzlu;
var
  i, k: integer;
begin
  if (nUzly <= kENDBDD) then
  begin // dosud nebylo zapsano zadne jmeno uzlu
    nUzly := kENDBDD + 1;
    i := nUzly;
    uzly[nUzly] := jmeno;
  end
  else begin // je treba prohledat uz zapsana jmena uzlu
    i := -1;
    for k := 0 to nUzly do
    begin
      if (jmeno = uzly[k]) then
      begin
        i := k;
        break;
      end;
    end;
    if (i < 0) then
    begin
      if (nUzly < maxJmen) then
      begin // do pole na konci prida nove jmeno
        nUzly := nUzly + 1;
        i := nUzly;
        uzly[nUzly] := jmeno;
      end
      else begin // prilis mnoho jmen
        kpreplneni := maxJmen;
        writeln(BDDMsg,'prilis mnoho uzlu')
      end;
    end;
  end;
  result := i;
end;

function kodPodminky;
var
  i, k: integer;
begin
  if (nBC <= 0) then
  begin // dosud nebyla zapsana ani jedna podminka
    nBC := 1;
    i := nBC;
    podminky[nBC] := jmeno;
  end
  else begin // je treba prohledat uz zapsana jmena podminek
    i := -1;
    for k := 0 to nBC do
    begin
      if (jmeno = podminky[k]) then
      begin
        i := k;
        break;
      end;
    end;
    if (i < 0) then
    begin
      if (nBC < maxJmen) then
      begin // do pole na konci prida nove jmeno
        nBC := nBC + 1;
        i := nBC;
        podminky[nBC] := jmeno;
      end
      else begin // prilis mnoho jmen
        kpreplneni := maxJmen - 9;
        writeln(BDDMsg,'prilis mnoho podminek');
      end;
    end;
  end;
  result := i;
end;
///******************************************************************

procedure ProchazeniUzly;
var
  i, kab, kk, ii, mUzel, ken1, ken2, mkonec, kStav : integer;
  konecTypuAno, negaceBC, zapsat: boolean;
begin
   // ***vrcholovy uzel (TOP), od nehoz se ma analyzovat BDD
  if (ktop = 0) then
  begin // vrcholovy uzel nebyl zadany na konci popisu BDD
    for i := kENDBDD + 1 to nUzly do
    begin
      if (jIn[i] = 0) then begin ktop := i; break; end;
    end;
  end;
  Writeln(BDDMsg, ' Prohledavani BDD zacne uzlem: ' + uzly0[ktop]);
  for i := 0 to maxJmen do jPlay[i] := 0;
  if (ktop = 0) then // chyba: neexistuje uzel, do nehoz nic nevstupuje
  begin
    status := 111;
    Writeln(BDDMsg,'chyba: nenalezen vrcholovy uzel');
  end
  else begin
      // ***od vrcholovho uzlu TOP se projdou vechny trasy dolu
      //    Zaznam, ktere uzly se pritom pouziji a kolikrat
      //    Test: stromova struktura BDD nebo nektery uzel ma > 1 vstup
    nZasobnik := 1; maxLevel := 1;
    jZasobnikTrasy[1] := -ktop;
    while nZasobnik > 0 do
    begin
      ii := jZasobnikTrasy[nZasobnik];
      if (ii < 0) then
      begin // *****test spoje pro pripad nesplneni podminky
        ii := -ii;
        jZasobnikTrasy[nZasobnik] := ii;
        jPlay[ii] := jPlay[ii] + 1; // priznak, ze timto uzlem BDD se proslo
        mUzel := jDef[ii]; // poradove cislo, kde je popis uzlu
        kk := jF[mUzel]; // kod spoje pri nesplneni podminky
        if ((kk > kENDBDD) and (jPlay[ii] = 1)) then
        begin // nasleduje nepouzity uzel - da se do zasobniku
          nZasobnik := nZasobnik + 1;
          maxLevel := max(nZasobnik, maxLevel);
          jZasobnikTrasy[nZasobnik] := -kk;
        end
        else begin
          if (jPlay[ii] > 1) then nZasobnik := nZasobnik - 1;
        end;
      end
      else begin // *****test spoje pri splneni podminky
        if (jPlay[ii] = 0) then jPlay[ii] := 1;
        mUzel := jDef[ii];
        kk := jT[mUzel]; // kod spoje pri splneni podminky
        if ((kk > kENDBDD) and (jPlay[ii] > 0)) then
        begin // nasleduje nepouzity uzel - da se do zasobniku
          jZasobnikTrasy[nZasobnik] := -kk;
        end
        else begin // nasleduje koncovy stav
          nZasobnik := nZasobnik - 1;
        end;
      end;
    end; //while
  end;
   //*** vyhleda pouzite uzly, secte jejich podminky
  konecTypuAno := (kend = 0) or ((kend > 0) and (kend < 5));
  if (konecTypuAno) then  Writeln(BDDMsg, ' hledaji se konce typu ANO')
  else Writeln(BDDMsg, ' hledaji se konce typu NE');
  ken1 := 4; ken2 := kENDBDD;
  maxPlay := 0; kStav := 0; mkonec := 0;
  for i := 0 to nuc do
  begin
    ii := jF[i];
    if ((ii > ken1) and (ii < ken2)) then jF[i] := 5;
    if ((ii > 0) and (ii <= ken1)) then jF[i] := 1;
    ii := jT[i];
    if ((ii > ken1) and (ii < ken2)) then jT[i] := 5;
    if ((ii > 0) and (ii <= ken1)) then jT[i] := 1;
  end;
  ii := 5;
  if konecTypuAno then ii := 1;
  if (bdebug) then Writeln(BDDMsg, 'Pouzite uzly: kod, jmeno ');
  for i := (kENDBDD + 1) to nUzly do
  begin
    if (jPlay[i] > 0) then
    begin
      maxPlay := maxPlay + 1;
      idef := jDef[i];
      if (bdebug) then Writeln(BDDMsg, inttostr(i) + ' =  ' + uzly0[i]);
      iBC := jBC[idef]; 
      if (jStavy[iBC] = 0) then kStav := kStav + 1;
      jStavy[iBC] := jStavy[iBC] + 1;
      if ((jF[idef] = ii) or (jT[idef] = ii)) then mkonec := mkonec + 1;
    end;
  end;
   // zda v dane casti BDD existuje zadany vystupni stav
  if ((mkonec = 0) or (kStav = 0)) then begin status := 202; exit; end;
   // poet slov pro ulozeni nenegovanych, resp. vsech podminek do soucinu
  iw := ((nBC - 1) div i32) + 1;
  iww := iw + iw;
  nSOP := 0;
   //  *vyhledani SOP, tj. vsechny trasy od TOP uzlu k danemu koncovemu stavu
  nZasobnik := 1;
  jZasobnikTrasy[1] := -ktop;
  while ((nZasobnik > 0) and (status = 0)) do
  begin
    ii := jZasobnikTrasy[nZasobnik];
    if (ii = 0) then
    begin
      nZasobnik := nZasobnik - 1;
    end
    else begin
      kab := Abs(ii);
      mUzel := jDef[kab]; // poradove cislo, kde je popis uzlu
      negaceBC := (ii < 0);
      if negaceBC then kk := jF[mUzel] else kk := jT[mUzel];
      if (kk > kENDBDD) then // test, zda nasleduje dalsi uzel
      begin // je to uzel - da se do zasobniku
        nZasobnik := nZasobnik + 1;
        jZasobnikTrasy[nZasobnik] := -kk;
      end
      else begin // doslo se na koncovy stav
        zapsat := ((kk = 1) and konecTypuAno) or ((kk = 5) and not (konecTypuAno));
        if (zapsat) then zapsaniTrasy;
        //if not (negaceBC) then kab := 0; // zda uzel ma vyrizene oba vystupy
        //jZasobnikTrasy[nZasobnik] := kab;
        if (negaceBC) then jZasobnikTrasy[nZasobnik] := kab
        else begin  // uzel uz ma vyrizene oba vystupy
          nZasobnik := nZasobnik - 1;
          // test, zda predchozi uzel ma take vyrizene oba vstupy
          while ((nZasobnik > 0) and (jZasobnikTrasy[nZasobnik] >= 0)) do
          nZasobnik := nZasobnik - 1;
          jZasobnikTrasy[nZasobnik] := abs(jZasobnikTrasy[nZasobnik]);
        end;
      end;
    end; // v zasobniku obsah ii byl nenulovy
  end; //while
end;

///*********************************************************************

procedure zapsaniTrasy;
var
  i, ii, kk, kPrvek, kBit, kneg: integer;
  neprazdna: boolean;
  buf : string;
begin
  // pomocne ladici vypisy
  if (bdebug) then
  begin
    Writeln(BDDMsg, ' ****trasa ' + inttostr(1 + nSOP div iww));
    buf :=  '';
    for i := 1 to nZasobnik do buf := buf + ' ' + inttostr(jZasobnikTrasy[i]);
    Writeln(BDDMsg, buf);
  end;
  if (nSOP + iww > maxSOP) then
  begin // ochrana proti preplneni pracovniho pole
    status := 200; exit;
  end;
   // vynulovani casti pole, kam se zakoduje novy soucin
  for i := nSOP + 1 to nSOP + iww do jSOP[i] := 0;
   // v jZasobnikTrasy jsou kody uzlu, ktere predstavuj 1 nalezeny soucin
  neprazdna := False;
  i := 0;
  while (i < nZasobnik) do
  begin
    i := i + 1;
    kk := jZasobnikTrasy[i];
    if (kk <> 0) then
    begin
      neprazdna := True;
      kneg := 0;
      if (kk < 0) then kneg := iw;
      kk := Abs(kk);
      kk := jDef[kk];
      iBC := jBC[kk];
      kPrvek := iBC div i32;
      kBit := iBC - kPrvek * i32;
      ii := nSOP + kneg + kPrvek + 1;
      jSOP[ii] := jSOP[ii] or maska[kBit];
    end;
  end; // while i
  for i := nSOP + 1 to nSOP + iw do
  begin // soucin se nezapise, je-li nulovy = je-li tu podminka i jeji negace
    if ((jSOP[i] and jSOP[i + iw]) <> 0) then neprazdna := False;
  end;
  if (neprazdna) then nSOP := nSOP + iww;
end;

///******************************************************************

procedure OdstraneniRedundantnichCasti;
var
  i, iBit, k, ki, kk, kSOP: integer;
  m, mp, mw, ma, mab, mm, mVel, mnvel, mxvel : integer;
  redundantni: boolean;
begin
   // for i:= 1 to nSOP do jPrac[i] := jSOP[i];
  kSOP := nSOP div iww; // kSOP = celkovy pocet soucinu
  if (kSOP <= 1) then exit;
   // spocita se, kolik ma kazdy soucin cinitelu
  mxVel := 0; mnVel := maxJmen;
  for kk := 1 to kSOP do // kk = poradove cislo soucinu
  begin
    mVel := 0; // sumace poctu bitu v soucinu
    k := (kk - 1) * iww;
    for i := 1 to iww do
    begin
      k := k + 1;
      m := jSOP[k];
      iBit := 0;
      while (m <> 0) do
      begin
        iBit := iBit + 1;
        if ((m and maska[iBit]) <> 0) then
        begin
          mVel := mVel + 1;
          m := m and komplementMaska[iBit];
        end;
      end;
    end;
    mxVel := Max(mxVel, mVel);
    mnVel := Min(mnVel, mVel);
    jVel[kk] := mVel;
  end;
   // v poli jPrac budou ulozeny souciny vzestupne podle velikosti
  mnVel := Max(mnVel, 1);
  nPrac := 0;
  for mm := mnVel to mxVel do
  begin
    for kk := 1 to kSOP do
      if (jVel[kk] = mm) then
      begin
        k := (kk - 1) * iww;
        for i := 1 to iww do
        begin
          k := k + 1;
          nPrac := nPrac + 1;
          jPrac[nPrac] := jSOP[k];
        end;
        jVel[kk] := -mm;
      end;
  end;
   // kazdy soucin se porovna se vsemi nasledujicimi, zda nasledujici je
   // logicky redundantni podle pravidla A+AB=A, kde A i B mohou byt souciny
  for i := 1 to kSOP do jVel[i] := 0;
  for kk := 1 to (kSOP - 1) do // cyklus pres vychozi soucin
  begin
    ki := (kk - 1) * iww;
    for mm := (kk + 1) to kSOP do // cyklus pres testovany stejny/delsi soucin
    if (jVel[mm] = 0) then
    begin
      redundantni := False;
      mp := (mm - 1) * iww;
      k := ki;
      for i := 1 to iww do // cyklus pres slova kodujici soucin
      begin
        k := k + 1;
        mp := mp + 1;
        ma := jPrac[k];
        mab := jPrac[mp];
        mw := ma or mab;
        if (mw = mab) then // je-li sjednoceni kratsiho a delsiho soucinu
        begin // totozne s delsim soucinem, ten delsi je zbytecny
          redundantni := True;
        end
        else begin
          redundantni := False;
          break;
        end;
      end; // i
      if (redundantni) then jVel[mm] := 1;
    end;  // mm
  end; // kk
        // do jSOP zapise neredundantni souciny s jPrac
  nSOP := 0;
  //mp := 0;
  for kk := 1 to kSOP do
    if (jVel[kk] = 0) then
    begin
      //mp := mp + 1;
      k := (kk - 1) * iww;
      for i := 1 to iww do
      begin
        k := k + 1;
        nSOP := nSOP + 1;
        jSOP[nSOP] := jPrac[k];
      end;
    end;
end;

///******************************************************************

procedure VypsatVystupy;
var
  i, iBit, k, kk, kSOP, m, mneg, mpom  : integer;
  bufLine, bufBC: string;
begin
   // cykl dekodovani jednotlivych soucinu ulozenych v poli jSOP
  kSOP := nSOP div iww; // kSOP = celkovy pocet soucinu
  for kk := 1 to kSOP do // kk = poradove cislo soucinu
  begin
    nPrac := 0; // poradove cislo logicke podminky v soucinu
    k := (kk - 1) * iww;
    for i := 1 to iw do
    begin
      k := k + 1;
      m := jSOP[k];
      mneg := jSOP[k + iw];
      iBit := 0;
      while (m <> 0) do
      begin
        iBit := iBit + 1;
        if ((m and maska[iBit]) <> 0) then
        begin
          nPrac := nPrac + 1;
          jPrac[nPrac] := (i - 1) * i32 + iBit;
          m := m and komplementMaska[iBit];
        end;
      end;
      iBit := 0;
      while (mneg <> 0) do
      begin
        iBit := iBit + 1;
        mpom := maska[iBit];
        mpom := mpom and mneg;
        if (mpom <> 0) then
        //if ((mneg and maska[iBit]) > 0) then
        begin
          nPrac := nPrac + 1;
          jPrac[nPrac] := -((i - 1) * i32 + iBit);
          mneg := mneg and komplementMaska[iBit];
        end;
      end;
    end; // cykl iw pres bity kodujici soucin
           // zapis poradoveho cisla a podmnek ze soucinu do textoveho souboru
    bufLine := ' #' + IntToStr(kk) + ': ';
    i := 1;
    while (i <= nPrac) do
    begin
      iBC := jPrac[i];
      if (iBC > 0) then
      begin // logicka podminka v nenegovanem tvaru
        bufBC := '  ' + Trim(podminky0[iBC]);
      end
      else begin // negace logicke podminky - vyznacono minusem pred ni
        iBC := -iBC;
        bufBC := ' -' + Trim(podminky0[iBC]);
      end;
      bufLine := bufLine + bufBC; // pripoji puvodni text podminky k radku
      if (Length(bufLine) > maxVystupZnaku) then
      begin // je-li radek moc dlouhy, zapise ho
        writeln(BDDMsg, bufLine); bufLine := '      ';
        //buf1 := '      ';
        //bufLine := buf0 + buf1; // zacne novy radek vypisu soucinu
      end;
      i := i + 1;
    end;
    writeln(BDDMsg, bufLine);
  end; // cykl kk pres log. souciny

end;

procedure TForm1.Button1Click(Sender: TObject);
begin
   //OtevriSouborPopisuBDD;
  NactiPopisBDD;
end;

procedure VystrahaSyntax;
begin
  MessageDlg('Formaln chyby v zadn BDD.'
    + #13 + #10 + 'Vpoet kon.', mtWarning, [mbOK], 0);
end;

procedure VystrahaPrazdnySoubor;
begin
  MessageDlg('Vstupni soubor je przdn.'
    + #13 + #10 + 'Vpoet kon.', mtWarning, [mbOK], 0);
end;

procedure VystrahaChybnaStruktura;
begin
  MessageDlg('Struktura BDD neumouje provst zadn.'
    + #13 + #10 + 'Vpoet kon.', mtWarning, [mbOK], 0);
end;

procedure VystrahaPreplneni;
begin
  MessageDlg('Pli velk BDD.'
    + #13 + #10 + 'Vpoet kon.', mtWarning, [mbOK], 0);
end;

procedure ZpravaVypocetUkoncen;
begin
  MessageDlg('Vpoet spn ukonen', mtCustom, [mbOK], 0);
end;
 
procedure TForm1.Button2Click(Sender: TObject);
begin
  form1.Close;
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
//////////////////

end;

procedure TForm1.Button5Click(Sender: TObject);
begin
 Memo1.Lines.LoadFromFile(vystupniSoubor);
 Memo1.Lines.Add(vystupniSoubor);
end;

procedure TForm1.Button3Click(Sender: TObject);
begin
 Memo1.Lines.LoadFromFile(vstupniSoubor);
 Memo1.Lines.Add(vstupniSoubor);
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  nazevFormulare:='bddplus';
  Form1.Caption := nazevFormulare;

end;

end.

